/*
******************************************************************************

   LibBuf extension API header file.

   Copyright (c) Wed Aug 29 2001 by: 
	Brian S. Julin		bri@tull.umassp.edu

  
   
This software may be redistributed under the terms of the GNU Library
(a.k.a. "Lesser") Public License -- see the file COPYING in the parent 
directory of the source code.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


******************************************************************************
*/

#include <stdio.h>
#include <string.h>

#include "config.h"
#include <ggi/internal/buf.h>
#include <ggi/internal/buf_debug.h>

#include <ggi/gg.h> /* We use LibGG to manage config files */


/* Static variables 
 */
static int 	 _BufLibIsUp       = 0;
static void	*_BufConfigHandle  = NULL;
static char      _Bufconfstub[512] = BUFCONFFILE;
static char     *_Bufconffile      = _Bufconfstub+BUFTAGLEN;

static ggiGA_resource_list reqdummy = NULL;

int _bufDebugSync = 0;
uint32 _bufDebugState = 0;



/* Extension ID. Defaulting to -1 should make segfault on abuse more likely...
   This is exported so that sublibs can use it.
 */
EXPORTVAR ggi_extid _BufID = -1;


#define SUBLIB_PREFIX   "BUFdl_"


/* Dummy function which returns -1
   We use this to reset the function pointers */
static int dummyfunc(void)
{
	BUFDPRINT_CORE("dummyfunc() of LibBUF called\n");
	return -1;
}	/* dummyfunc */


/*-* Reset all function pointers to dummyfunc */
static void clearfuncs(bufpriv *priv)
{
	priv->strip	= (void *) dummyfunc;
	priv->_untie	= (void *) dummyfunc;
	priv->_clad	= (void *) dummyfunc;
	priv->_tie	= (void *) dummyfunc;
}	/* clearfuncs */


/* This function is called by LibGGI when the API-list changes */
static int changed(ggi_visual_t vis, int whatchanged)
{
	BUFDPRINT_LIBS("changed(%p, 0x%x) of LibBUF called\n",
		       vis, whatchanged);

	switch (whatchanged) {
	case GGI_CHG_APILIST: {
		int i;
		const char *fname;
		ggi_dlhandle *lib;
		char api[GGI_MAX_APILEN], args[GGI_MAX_APILEN];
		bufpriv *priv;

		priv = LIBGGI_BUFEXT(vis);

		clearfuncs(priv);

		/* We use the old getapi hook, avoiding recursion. */
		for (i=0; priv->getapi(vis, i, api, args) == 0; i++) {
			strcat(api,"-buf");
			fname = ggMatchConfig(_BufConfigHandle, api, NULL);
			if (fname == NULL) {
				/* No special implementation for this sublib */
				continue;
			}	/* if */
			lib = ggiExtensionLoadDL(vis, fname, args, NULL, 
						 SUBLIB_PREFIX);
		}
	}
	break;
	}	/* switch */
                                
	return 0;
}	/* changed */



/* Initialize the extension
 */
int ggiBufInit(void)
{
	int err;
	char *str;

	_BufLibIsUp++;
	if (_BufLibIsUp > 1) return 0; /* Initialize only at first call. */

	str = getenv("BUF_DEBUG");
	if (str != NULL) {
		_bufDebugState = atoi(str);
		BUFDPRINT_CORE("Debugging=%d\n", _bufDebugState);
	}	/* if */

	str = getenv("BUF_DEBUGSYNC");
	if (str != NULL) {
		_bufDebugSync = 1;
	}	/* if */

	err = ggLoadConfig(_Bufconffile, &_BufConfigHandle);
	if (err != GGI_OK) {
		fprintf(stderr,"LibBuf: couldn't open %s\n",
			_Bufconffile);
		_BufLibIsUp--;
		return err;
	}	/* if */

	_BufID = ggiExtensionRegister("BUF", sizeof(bufpriv), changed);

	if (_BufID < 0) {
		fprintf(stderr,
			"LibBuf: failed to register as extension.\n");
		_BufLibIsUp--;
		ggFreeConfig(_BufConfigHandle);
		return _BufID;
	}	/* if */

	BUFDPRINT_CORE("LibBUF Initialized.  Now Initializing LibGAlloc\n");

	if (ggiGAInit() != GGI_OK) {
		BUFDPRINT_CORE("Failure when initializing LibGAlloc\n");
		ggiBufExit();
		return GGI_EFATAL;
	}

	return 0;
}	/* ggiBufInit */


/* Deinitialize the extension
 */
int ggiBufExit(void)
{
	int rc;

	if (!_BufLibIsUp) return -1;

	if (_BufLibIsUp > 1) {
		BUFDPRINT_CORE("LibBUF exiting this instance.\n");
		/* Exit only at last call */
		_BufLibIsUp--;
		return 0;
	}	/* if */

	BUFDPRINT_CORE("LibBUF exiting last instance.\n");
	rc = ggiExtensionUnregister(_BufID);
	ggFreeConfig(_BufConfigHandle);

	_BufLibIsUp = 0;

	ggiGAExit();

	return rc;
}	/* ggiBufExit */


/* Attach the extension to a visual
 */
int ggiBufAttach(ggi_visual_t vis, ggiGA_resource_list *reqlist)
{
	int rc;

	BUFDPRINT_CORE("ggiBufAttach(%p) called\n", vis);

	rc = ggiExtensionAttach(vis, _BufID);
	if (rc < 0) {
		BUFDPRINT_CORE("Failed to attach myself to this visual.\n");
		goto err0;
	}	/* if */

	if (rc != 0) return rc;

	rc = ggiGAAttach(vis);
	if (rc < 0) {
		BUFDPRINT_CORE("LibGAlloc Failed to attach to this visual.\n");
		goto err1;
	}	/* if */
	rc = 0;
	
	/* We are actually creating the primary instance. */
	memset(LIBGGI_BUFEXT(vis), 0, sizeof(bufpriv));
	LIBGGI_BUFEXT(vis)->priv = NULL;
	
	/*-* Apply initial values */
	LIBGGI_BUFEXT(vis)->is_clad = 0;
	LIBGGI_BUFEXT(vis)->old_opdraw = NULL;
	LIBGGI_BUFEXT(vis)->abuf = NULL;
	LIBGGI_BUFEXT(vis)->zbuf = NULL;
	LIBGGI_BUFEXT(vis)->getapi = vis->opdisplay->getapi;
	LIBGGI_BUFEXT(vis)->cfghand = _BufConfigHandle;
	LIBGGI_BUFEXT(vis)->achannel = 0;
	
	if (reqlist == NULL) {
		LIBGGI_BUFEXT(vis)->reqlist = &reqdummy;
	} else {
		LIBGGI_BUFEXT(vis)->reqlist = reqlist;
	}	/* if */

	/* Now fake an "API change" so the right libs get loaded */
	changed(vis, GGI_CHG_APILIST);

	return rc;

err1:
	ggiExtensionDetach(vis, _BufID);
err0:
	return rc;
}	/* ggiBufAttach */


/* Detach the extension from a visual
 */
int ggiBufDetach(ggi_visual_t vis)
{
	int rc;
	ggiGA_resource_list list = NULL;

	bufpriv *priv = NULL;

	priv = LIBGGI_BUFEXT(vis)->priv;

	BUFDPRINT_CORE("ggiBufDetach(%p) called\n", vis);

	rc = ggiExtensionDetach(vis, _BufID);
	if (rc != 0) return rc;

	rc = ggiGAGet(vis, &list);
	LIB_ASSERT(rc == GALLOC_OK);

	if (priv != NULL) {
		if (priv->old_opdraw != NULL) {
			if (priv->is_clad) {
				/* Restore the visual's drawing primitives. */
				memcpy(vis->opdraw, priv->old_opdraw, 
				       sizeof(struct ggi_visual_opdraw));
			}	/* if */

			free(priv->old_opdraw);
		}	/* if */

		free(priv);
	}	/* if */

	ggiGADetach(vis);

	return rc;
}	/* ggiBufDetach */


/* This should be included by exactly _one_ file in each extension library. */
#include <ggi/internal/ggilibinit.h>
